"""
Log parsing exceptions
"""


class ParsingError(Exception):
	"""
	Generic exception that happens during log parsing.
	"""
	pass


class RegexParsingError(ParsingError):
	pass


class ExporterError(Exception):
	"""
	Generic exception that happens during PacketTree export.
	"""
	pass


class MissingPlayerData(RuntimeError):
	"""
	Raised when it is not possible to reverse a LazyPlayer instance.
	"""
	pass



from collections.abc import Iterable
from typing import IO, List, Optional, Tuple, Union

from hearthstone.enums import GameTag

from hslog import tokens
from hslog.exceptions import RegexParsingError
from hslog.utils import parse_tag


# List of TAG_CHANGE tags to discard

BLACKLISTED_TAGS = [
    "EXHAUSTED",
    "FACTION",
    "HIDE_COST",
    "HIDE_WATERMARK",
    "NUM_ATTACKS_THIS_TURN",
    "NUM_TURNS_IN_PLAY",
    "TECH_LEVEL",
]

# List of FULL_ENTITY tags to discard

BLACKLISTED_FULL_ENTITY_TAGS = [
    GameTag.BACON_ACTION_CARD,
    GameTag.BATTLECRY,
    GameTag.CARDRACE,
    GameTag.COST,
    GameTag.EXHAUSTED,
    GameTag.FACTION,
    GameTag.GAME_MODE_BUTTON_SLOT,
    GameTag.HIDE_WATERMARK,
    GameTag.IS_BACON_POOL_MINION,
    GameTag.MODULAR,
    GameTag.MOVE_MINION_HOVER_TARGET_SLOT,
    GameTag.OVERKILL,
    GameTag.RARITY,
    GameTag.SPAWN_TIME_COUNT,
    GameTag.SUPPRESS_ALL_SUMMON_VO,
    GameTag.TAG_LAST_KNOWN_COST_IN_HAND,
    GameTag.TAG_SCRIPT_DATA_NUM_1,
    GameTag.TAG_SCRIPT_DATA_NUM_2,
    GameTag.TECH_LEVEL,
    GameTag.TRIGGER_VISUAL
]

# List of SHOW_ENTITY tags to discard (in addition to BLACKLISTED_FULL_ENTITY_TAGS above)

BLACKLISTED_SHOW_ENTITY_TAGS = [
    GameTag.ENCHANTMENT_INVISIBLE
]


class Buffer:
    """Represents a sequence of buffered log lines that might be emitted or skipped.

    Buffers can be nested, such that the collection of buffered lines includes both strings
    and other Buffer instances. Output skipping should be applied hierarchically such that
    `should_skip` flag set on a "parent" Buffer should cause lines in any nested Buffers to
    be skipped, regardless of the value of that flag for those instances.
    """

    def __init__(self, buffer_type: str, subtype: str, parent: Optional["Buffer"] = None):
        """Ctor.

        :param buffer_type: the type of log element being buffered, e.g. "BLOCK_START"
        :param subtype: the subtype of log element being buffered, e.g. "ATTACK"
        :param: the parent Buffer instance, if any
        """

        self.buffer_type = buffer_type
        self.subtype = subtype
        self.buffer: List[Union[Buffer, str]] = []
        self.parent: Optional[Buffer] = parent
        self.should_skip = False


class BattlegroundsLogFilter(Iterable):
    """Iterable implementation that discards Battlegrounds log lines not needed for Tier7.

    Wrap this filter around the file-like object for a Battlegrounds log before passing it
    to the regular Hearthstone LogParser.

    The following log-skipping strategies are employed:

    - ATTACK blocks for minions that do not include interesting TRIGGER subblocks
        (such as those for DEATHRATTLEs) are discarded; ATTACK blocks for heroes will cause
        the next 4 blocks to be preserved, consistent with the logic for refreshing the
        `battlegrounds_combat_snapshot` materialized view
    - DEATHS blocks with no subblocks are discarded
    - All options messages (from DebugPrintOptions) are discarded
    - Blacklisted and unknown tags for FULL_ENTITY and SHOW_ENTITY messages are discarded
    - TAG_CHANGES containing blacklisted and unknown tags are discarded
    - All TAG_CHANGES that are part of a TRIGGER block that include a BOARD_VISUAL_STATE=2
        tag change are discarded
    - All TAG_CHANGES that precede a BOARD_VISUAL_STATE=1 tag change in a TRIGGER block are
        discarded
    """

    def __init__(self, fp: IO, show_suppressed_lines: bool = False):
        """Ctor.

        :param fp: the file-like object to be filtered
        :param show_suppressed_lines: whether to hide or show suppressed lines; when shown,
            suppressed lines are prefixed with "X: "
        """

        self._fp = fp
        self._preserve_block_counter = 0
        self._show_suppressed_lines = show_suppressed_lines

        self._current_buffer = None
        self._flushed_lines = []

        self.num_lines_read = 0
        self.num_lines_emitted = 0

    # Returns True if there's a current buffer which is buffering a FULL_ENTITY or
    # SHOW_ENTITY sequence of messages, False otherwise.

    def _buffering_entity(self) -> bool:
        return (
            self._current_buffer is not None and
            self._current_buffer.buffer_type in ["FULL_ENTITY", "SHOW_ENTITY"]
        )

    # Adds a line to the current buffer if it exists or to the outbound sequence of
    # "flushed" lines otherwise.

    def _emit_line(self, line):
        if self._current_buffer:
            self._current_buffer.buffer.append(line)
        else:
            self._flushed_lines.append(line)

    # Closes the current buffer sets the new current buffer to its parent. If the current
    # buffer had no parent, all non-skipped lines are recursively flushed to the flushed
    # lines list to be subsequently returned.

    def _end_buffer(self):
        if self._current_buffer:
            if self._current_buffer.parent is not None:
                self._current_buffer = self._current_buffer.parent
            else:
                buffer = self._current_buffer
                self._current_buffer = None
                self._flush_buffered_packets(buffer, should_skip=buffer.should_skip)

    # Recursively flushes buffered lines to the flushd lines list. If "should_skip" is True,
    # all lines in the current buffer and any nested buffers are suppressed, regardless of
    # their local "should_skip" flag.

    def _flush_buffered_packets(self, buffer: Buffer, should_skip: bool = False):
        for buffered_item in buffer.buffer:
            if isinstance(buffered_item, Buffer):
                eff_should_skip = should_skip or buffered_item.should_skip
                self._flush_buffered_packets(buffered_item, should_skip=eff_should_skip)
            else:
                if should_skip:
                    if self._show_suppressed_lines:
                        self._flushed_lines.append("X: " + buffered_item)
                else:
                    self._flushed_lines.append(buffered_item)

    # Given the speciifed "opcode" (e.g., "BLOCK_START") and the following remainder of the
    # line, parse and return a Tuple of the block type (e.g., "ATTACK") and the card id of
    # the related entity.

    def _get_block_type_and_card_id(self, opcode, data) -> Tuple[str, str]:
        _index = None
        _effectid, _effectindex = None, None
        _suboption, _trigger_keyword = None, None
        if " SubOption=" in data:
            if " TriggerKeyword=" in data:
                sre = tokens.BLOCK_START_20457_TRIGGER_KEYWORD_RE.match(data)
                if sre is None:
                    raise RegexParsingError(data)
                block_type, entity, _effectid, _effectindex, target, _suboption, \
                    _trigger_keyword = sre.groups()
            else:
                sre = tokens.BLOCK_START_20457_RE.match(data)
                if sre is None:
                    raise RegexParsingError(data)
                block_type, entity, _effectid, _effectindex, target, \
                    _suboption = sre.groups()
        else:
            if opcode == "ACTION_START":
                sre = tokens.ACTION_START_RE.match(data)
            else:
                sre = tokens.BLOCK_START_12051_RE.match(data)

            if sre is None:
                sre = tokens.ACTION_START_OLD_RE.match(data)
                if not sre:
                    raise RegexParsingError(data)
                entity, block_type, _index, target = sre.groups()
            else:
                block_type, entity, _effectid, _effectindex, target = sre.groups()

        return block_type, self._get_card_id(entity)

    # Return the card id (if present) for the specified "entity string."

    @staticmethod
    def _get_card_id(entity: str) -> Optional[str]:
        card_id_index = entity.find("cardId=")
        if card_id_index < 0:
            return None

        card_id_index += 7
        following_space_index = entity.find(" ", card_id_index)
        return entity[card_id_index:following_space_index] \
            if following_space_index > card_id_index \
            else entity[card_id_index:]

    # Return a tuple of tag name and tag value from the specified log line.

    @staticmethod
    def _get_tag_change_tag_and_value(data) -> Tuple[str, str]:
        sre = tokens.TAG_CHANGE_RE.match(data)
        _e, tag, value, _def_change = sre.groups()
        return tag, value

    # Buffer or emit lines related to the specified BLOCK_END message.

    def _handle_block_end(self, line: str):

        # If we're buffering tag messages for a FULL_ENTITY / SHOW_ENTITY, flush them.

        if self._buffering_entity():
            self._end_buffer()

        self._emit_line(line)

        if self._current_buffer:
            self._end_buffer()

    # Buffer or emit lines related to the specified BLOCK_START message.

    def _handle_block_start(self, msg: str, line: str):

        # If we're buffering tag messages for a FULL_ENTITY / SHOW_ENTITY, flush them.

        if self._buffering_entity():
            self._end_buffer()

        block_type, card_id = self._get_block_type_and_card_id("BLOCK_START", msg)

        if self._current_buffer:

            # If we're already buffering an ATTACK block, check and see if this new block
            # is a TRIGGER related to something we care about, like a deathrattle. If so,
            # we should keep the outer block "alive."

            if self._current_buffer.subtype == "ATTACK":
                if block_type == "TRIGGER" and card_id != "TB_BaconShop_8P_PlayerE":
                    self._current_buffer.should_skip = False

            # If we're already buffering a DEATHS block, then any nested block (like this
            # one) should serve to preserve it.

            elif self._current_buffer.subtype == "DEATHS":
                self._current_buffer.should_skip = False

        if block_type == "ATTACK":

            # If the hero is attacking, preserve this block as well as the next 4 blocks -
            # there's a calculation in the materialized view update logic for
            # `battlegrounds_combat_snapshot` that expects a roughly 4-block gap between
            # a hero attacking and a hero death notification.

            if self._is_hero(card_id):
                self._preserve_block_counter = 4
            else:

                # Minion attacks aren't interesting unless they trigger something
                # interesting (see above)

                self._start_new_buffer("BLOCK", block_type)
                self._current_buffer.should_skip = True

        # Deaths aren't interesting unless they contain a nested block (see above)

        elif block_type == "DEATHS":
            self._start_new_buffer("BLOCK", block_type)
            self._current_buffer.should_skip = True

        # A TRIGGER block from the "8 player enchantment" might indicate a board state
        # reset that contains skippable stuff - see TAG_CHANGE logic below.

        elif block_type == "TRIGGER":
            if card_id == "TB_BaconShop_8P_PlayerE":
                self._start_new_buffer("BLOCK", block_type)

        if self._preserve_block_counter > 0:
            self._preserve_block_counter -= 1
            if self._current_buffer:
                self._current_buffer.should_skip = False

        self._emit_line(line)

    # Buffer or emit lines related to the specified FULL_ENTITY or SHOW_ENTITY message.

    def _handle_entity(self, opcode: str, line: str):

        # If we're buffering tag messages for a previous FULL_ENTITY / SHOW_ENTITY, flush
        # them.

        if self._buffering_entity():
            self._end_buffer()

        self._start_new_buffer(opcode, "")
        self._emit_line(line)

    # Buffer or emit lines related to the "initial atgs" attached to a previous FULL_ENTITY
    # or SHOW_ENTITY message.

    def _handle_entity_tag(self, msg: str, line: str):
        tag, value = self._parse_initial_tag(msg)

        if self._buffering_entity():
            if self._current_buffer.buffer_type == "FULL_ENTITY":
                if tag in BLACKLISTED_FULL_ENTITY_TAGS:
                    self._start_new_buffer("__ENTITY_TAG", "")
                    self._current_buffer.should_skip = True
                    self._emit_line(line)
                    self._end_buffer()
                else:
                    self._emit_line(line)
            elif self._current_buffer.buffer_type == "SHOW_ENTITY":

                # Filter out all tags that we blacklist for FULL_ENTITY as well

                if (
                        tag in BLACKLISTED_FULL_ENTITY_TAGS or
                        tag in BLACKLISTED_SHOW_ENTITY_TAGS
                ):
                    self._start_new_buffer("__ENTITY_TAG", "")
                    self._current_buffer.should_skip = True
                    self._emit_line(line)
                    self._end_buffer()
                else:
                    self._emit_line(line)
            else:
                self._emit_line(line)
        else:
            self._emit_line(line)

    # Buffer or emit lines related to the specified TAG_CHANGE message.

    def _handle_tag_change(self, msg: str, line: str):

        # If we're buffering tag messages for a FULL_ENTITY / SHOW_ENTITY, flush them.

        if self._buffering_entity():
            self._end_buffer()

        tag, value = self._get_tag_change_tag_and_value(msg)

        # BOARD_VISUAL_STATE changes indicate that Battlegrounds is manipulating the shop.
        # If we're setting it to "1" we want to discard all previous TAG_CHANGEs that were
        # part of the same enclosing TRIGGER block. If we're setting it to "2" we want to
        # discard those TAG_CHANGEs as well as all subsequent TAG_CHANGEs.

        if (
                self._current_buffer and
                self._current_buffer.subtype == "TRIGGER" and
                tag == "BOARD_VISUAL_STATE"
        ):
            self._current_buffer.buffer.append(line)

            if value == "1" or value == "2":
                for i in range(len(self._current_buffer.buffer)):
                    buffered_item = self._current_buffer.buffer[i]
                    if (
                            isinstance(buffered_item, str) and
                            "TAG_CHANGE" in buffered_item
                    ):
                        buf = Buffer("TAG_CHANGE", "", parent=self._current_buffer)
                        buf.buffer.append(buffered_item)
                        buf.should_skip = True
                        self._current_buffer.buffer[i] = buf

                if value == "2":
                    setattr(self._current_buffer, "skip_tag_changes", True)

                return

        if (
                tag.isdigit() or
                tag in BLACKLISTED_TAGS or
                (
                    self._current_buffer is not None and
                    hasattr(self._current_buffer, "skip_tag_changes")
                )
        ):
            self._start_new_buffer("TAG_CHANGE", "")
            self._current_buffer.should_skip = True
            self._emit_line(line)
            self._end_buffer()
        else:
            self._emit_line(line)

    # Buffer or emit lines related to the specified option message.

    def _handle_options(self, level: str, ts: str, msg: str, line: str):
        if msg.startswith("id="):

            # This is the log filtering version of the "options hack" that the parser uses
            # to terminate dangling blocks for BGS games. Since we're suppressing options
            # lines, the parser won't know that it needs to use the hack, so instead we'll
            # need to emit a synthetic "BLOCK_END."

            block_end = f"{level} {ts} GameState.DebugPrintPower() - BLOCK_END\n"
            self._emit_line(block_end)
            self._end_buffer()

        if self._show_suppressed_lines:
            suppressed_line = "X: " + line
            self._flushed_lines.append(suppressed_line)

    # Predicate for detecting whether a specified card id is a BGS hero; this is a naive
    # implementation that may not work consistently!

    @staticmethod
    def _is_hero(card_id: str) -> bool:
        return card_id and "HERO" in card_id

    # Parse and return a tuple of tag and value for the specified log message corresponding
    # to the initial tags for a FULL_ENTITY or SHOW_ENTITY message.

    @staticmethod
    def _parse_initial_tag(data: str) -> Tuple[GameTag, str]:
        sre = tokens.TAG_VALUE_RE.match(data)
        if not sre:
            raise RegexParsingError(data)
        tag, value = sre.groups()
        return parse_tag(tag, value)

    # Start a new buffer with the specified buffer type and subtype and set it to be the new
    # "current buffer." Parent pointer and buffer nesting are updated as part of this.

    def _start_new_buffer(self, buffer_type: str, subtype: str):
        new_buffer = Buffer(buffer_type, subtype, parent=self._current_buffer)

        if self._current_buffer:
            self._current_buffer.buffer.append(new_buffer)

        self._current_buffer = new_buffer

    def __iter__(self):
        return self

    def __next__(self):
        while True:

            # First, return any lines that have been flushed to the flushed line list.

            if self._flushed_lines:
                self.num_lines_emitted += 1
                return self._flushed_lines.pop(0)

            line = self._fp.readline()
            if line == "":
                raise StopIteration()

            self.num_lines_read += 1

            sre = tokens.TIMESTAMP_RE.match(line)
            if not sre:
                raise RegexParsingError(line)

            level, ts, line_rest = sre.groups()
            if line_rest.startswith(tokens.SPECTATOR_MODE_TOKEN):
                self._emit_line(line)
                continue

            sre = tokens.POWERLOG_LINE_RE.match(line_rest)
            if not sre:
                self._emit_line(line)
                continue

            method, msg = sre.groups()
            msg = msg.strip()

            if method == "GameState.DebugPrintPower":
                opcode = msg.split()[0]

                if opcode == "BLOCK_START":
                    self._handle_block_start(msg, line)
                elif opcode == "BLOCK_END":
                    self._handle_block_end(line)
                elif opcode in ["FULL_ENTITY", "SHOW_ENTITY"]:
                    self._handle_entity(opcode, line)
                elif opcode.startswith("tag="):
                    self._handle_entity_tag(msg, line)
                elif opcode == "TAG_CHANGE":
                    self._handle_tag_change(msg, line)
                else:
                    self._emit_line(line)

            elif method == "GameState.DebugPrintOptions":
                self._handle_options(level, ts, msg, line)
            else:
                self._emit_line(line)



import logging
from datetime import datetime, timedelta
from typing import Any, Callable, Dict, Optional, Union

from aniso8601 import parse_time
from hearthstone.enums import (
	BlockType, ChoiceType, FormatType, GameTag, GameType,
	MetaDataType, Mulligan, OptionType, PlayReq, PowerType
)

from . import packets, tokens
from .exceptions import ParsingError, RegexParsingError
from .packets import (
	Block, Choices, ChosenEntities, CreateGame,
	MetaData, Packet, PacketTree, SendChoices, SubSpell
)
from .player import PlayerManager, PlayerReference, coerce_to_entity_id
from .utils import parse_enum, parse_tag


class ParsingState:

	def __init__(self):
		self.current_block: Optional[Union[Packet, PacketTree]] = None
		self.game_meta = {}
		self.games = []
		self.manager = PlayerManager()
		self.mulligan_choices: Dict[int, Choices] = {}
		self.packet_tree: Optional[PacketTree] = None
		self.spectating_first_player = False
		self.spectating_second_player = False

		self.choice_packet: Optional[Choices] = None
		self.chosen_packet: Optional[ChosenEntities] = None
		self.chosen_packet_count: int = 0
		self.entity_packet: Optional[Packet] = None
		self.game_packet: Optional[CreateGame] = None
		self.metadata_node: Optional[MetaData] = None
		self.send_choice_packet: Optional[SendChoices] = None

	def block_end(self, ts):
		if not self.current_block.parent:
			logging.warning("[%s] Orphaned BLOCK_END detected", ts)
			return self.current_block

		if isinstance(self.current_block, Block):
			self.current_block.end()

		block = self.current_block
		self.current_block = self.current_block.parent
		return block

	def _register_player_name_mulligan(self, player: PlayerReference, packet: Choices):
		"""
		Attempt to register player names by looking at Mulligan choice packets.
		In Hearthstone 6.0+, registering a player name using tag changes is not
		available as early as before. That means games conceded at Mulligan no
		longer have player names.
		This technique uses the cards offered in Mulligan instead, registering
		the name of the packet's entity with the card's controller as PlayerID.
		"""
		if player.entity_id:
			# The player is already registered, ignore.
			return player
		if not player.name:
			# If we don't have the player name, we can't use this at all
			return player
		for choice in packet.choices:
			player_id = self.manager.get_controller_by_entity_id(choice)
			if player_id is None:
				raise ParsingError("Unknown entity ID in choice: %r" % choice)

			player = self.manager.create_or_update_player(
				name=player.name,
				player_id=player_id
			)

			packet.entity = player.entity_id

		return player

	def flush(self):
		if self.metadata_node:
			self.metadata_node = None

		if self.choice_packet:
			if self.choice_packet.type == ChoiceType.MULLIGAN:
				if isinstance(self.choice_packet.entity, PlayerReference):
					player = self._register_player_name_mulligan(
						self.choice_packet.entity,
						self.choice_packet
					)
					self.mulligan_choices[player.player_id] = self.choice_packet
			self.choice_packet = None

		self.chosen_packet = None

		if self.send_choice_packet:
			self.send_choice_packet = None

	def parse_entity_id(self, entity: str) -> int:
		if entity.isdigit():
			return int(entity)

		if entity == tokens.GAME_ENTITY:
			return int(self.game_packet.entity)

		sre = tokens.ENTITY_RE.match(entity)
		if sre:
			entity_id = sre.groups()[0]
			return int(entity_id)

	def parse_entity_or_player(self, entity) -> Union[None, int, PlayerReference]:
		if entity == "-1":
			return

		entity_id = self.parse_entity_id(entity)
		if entity_id is None:
			# Only case where an id is None is if it's a Player name
			return self.manager.create_or_update_player(name=entity)
		return entity_id

	def register_game(self, _ts: int, entity_id: int):
		# Use the timestamp from CREATE_GAME because it's earlier
		ts = self.packet_tree.ts
		self.game_packet = self.entity_packet = packets.CreateGame(ts, entity_id)
		self.register_packet(self.game_packet)
		return self.game_packet

	def register_packet(self, packet: Packet, node=None):
		if node is None:
			node = self.current_block.packets
		node.append(packet)
		self.packet_tree.packet_counter += 1
		packet.packet_id = self.packet_tree.packet_counter

	def register_player(self, ts, entity_id: int, player_id: int, hi: int, lo: int):
		hi = int(hi)
		lo = int(lo)
		lazy_player = self.manager.create_or_update_player(
			entity_id=entity_id,
			player_id=player_id,
			is_ai=lo == 0
		)
		self.entity_packet = packets.CreateGame.Player(ts, lazy_player, player_id, hi, lo)
		self.register_packet(self.entity_packet, node=self.game_packet.players)
		return lazy_player

	@property
	def spectator_mode(self):
		return self.spectating_first_player or self.spectating_second_player

	def set_spectating(self, first, second=None):
		self.spectating_first_player = first
		if second is not None:
			self.spectating_second_player = second


def parse_initial_tag(data):
	"""
	Parse \a data, a line formatted as tag=FOO value=BAR
	Returns the values as int.
	"""
	sre = tokens.TAG_VALUE_RE.match(data)
	if not sre:
		raise RegexParsingError(data)
	tag, value = sre.groups()
	return parse_tag(tag, value)


def clean_option_errors(error, error_param):
	"""
	As of 8.0.0.18336, all option packets are accompanied by an error and an
	errorParam argument.
	This function turns both into their respective types.
	"""

	if error == "NONE":
		error = None
	else:
		error = parse_enum(PlayReq, error)

	if not error_param:
		error_param = None
	else:
		error_param = int(error_param)

	return error, error_param


class HandlerBase:
	def __init__(self):
		self._game_state_processor = "GameState"

	def parse_method(self, m):
		return "%s.%s" % (self._game_state_processor, m)

	def find_callback(self, method: str) -> Callable[[ParsingState, int, Any], Any]:
		raise NotImplementedError()


class PowerHandler(HandlerBase):
	def __init__(self):
		super().__init__()

		self._creating_game = False

	@staticmethod
	def _check_for_mulligan_hack(ps: ParsingState, ts, tag, value):

		# Old game logs didn't handle asynchronous mulligans properly.
		# If we're missing an ACTION_END packet after the mulligan SendChoices,
		# we just close it out manually.

		if tag == GameTag.MULLIGAN_STATE and value == Mulligan.DEALING:
			assert ps.current_block
			if isinstance(ps.current_block, packets.Block):
				logging.warning("[%s] Broken mulligan nesting. Working around...", ts)
				ps.block_end(ts)

	def find_callback(self, method):
		if method == self.parse_method("DebugPrintPower"):
			return self.handle_data
		elif method == self.parse_method("DebugPrintGame"):
			return self.handle_game

	@staticmethod
	def handle_game(ps: ParsingState, _ts, data):
		if data.startswith("PlayerID="):
			sre = tokens.GAME_PLAYER_META.match(data)

			if not sre:
				raise RegexParsingError(data)

			player_id, player_name = sre.groups()
			ps.manager.create_or_update_player(name=player_name, player_id=int(player_id))
		else:
			key, value = data.split("=")
			key = key.strip()
			value = value.strip()
			if key == "GameType":
				value = parse_enum(GameType, value)
				ps.manager._game_type = value
			elif key == "FormatType":
				value = parse_enum(FormatType, value)
			else:
				value = int(value)

			ps.game_meta[key] = value

	def handle_data(self, ps: ParsingState, ts, data):
		opcode = data.split()[0]

		if opcode == "ERROR:":
			# Line error... skip
			return

		if opcode in PowerType.__members__:
			return self.handle_power(ps, ts, opcode, data)

		if opcode == "GameEntity":
			ps.flush()
			self._creating_game = True
			sre = tokens.GAME_ENTITY_RE.match(data)
			if not sre:
				raise RegexParsingError(data)

			entity_id_str, = sre.groups()
			ps.register_game(ts, int(entity_id_str))
		elif opcode == "Player":
			ps.flush()
			sre = tokens.PLAYER_ENTITY_RE.match(data)
			if not sre:
				raise RegexParsingError(data)

			entity_id_str, player_id_str, hi_str, lo_str = sre.groups()

			ps.register_player(
				ts,
				int(entity_id_str),
				int(player_id_str),
				int(hi_str),
				int(lo_str)
			)
		elif opcode.startswith("tag="):
			tag, value = parse_initial_tag(data)

			assert hasattr(ps.entity_packet, "tags")
			ps.entity_packet.tags.append((tag, value))  # noqa

			if tag == GameTag.CONTROLLER:

				# We need to know entity controllers for player name registration

				assert hasattr(ps.entity_packet, "entity")
				entity_id = coerce_to_entity_id(ps.entity_packet.entity)  # noqa
				ps.manager.register_controller(int(entity_id), int(value))

		elif opcode.startswith("Info["):
			if not ps.metadata_node:
				logging.warning("[%s] Metadata Info outside of META_DATA: %r", ts, data)
				return
			sre = tokens.METADATA_INFO_RE.match(data)
			if not sre:
				raise RegexParsingError(data)
			idx, entity = sre.groups()
			entity = ps.parse_entity_or_player(entity)
			ps.metadata_node.info.append(entity)
		elif opcode == "Source":
			sre = tokens.SUB_SPELL_START_SOURCE_RE.match(data)
			if not sre:
				raise RegexParsingError(data)
			entity, = sre.groups()
			entity = ps.parse_entity_or_player(entity)
			if not isinstance(ps.current_block, packets.SubSpell):
				logging.warning("[%s] SubSpell Source outside of SUB_SPELL: %r", ts, data)
				return
			ps.current_block.source = entity
		elif opcode.startswith("Targets["):
			sre = tokens.SUB_SPELL_START_TARGETS_RE.match(data)
			if not sre:
				raise RegexParsingError(data)
			idx, entity = sre.groups()
			entity = ps.parse_entity_or_player(entity)
			if not isinstance(ps.current_block, packets.SubSpell):
				logging.warning("[%s] SubSpell Target outside of SUB_SPELL: %r", ts, data)
				return
			ps.current_block.targets.append(entity)
		else:
			raise NotImplementedError(data)

	def handle_power(self, ps: ParsingState, ts, opcode, data):
		ps.flush()

		if opcode == "CREATE_GAME":
			regex, callback = tokens.CREATE_GAME_RE, self.create_game
		elif opcode in ("ACTION_START", "BLOCK_START"):
			index = None
			effectid, effectindex = None, None
			suboption, trigger_keyword = None, None
			if " SubOption=" in data:
				if " TriggerKeyword=" in data:
					sre = tokens.BLOCK_START_20457_TRIGGER_KEYWORD_RE.match(data)
					if sre is None:
						raise RegexParsingError(data)
					(
						block_type,
						entity,
						effectid,
						effectindex,
						target,
						suboption,
						trigger_keyword
					) = sre.groups()
				else:
					sre = tokens.BLOCK_START_20457_RE.match(data)
					if sre is None:
						raise RegexParsingError(data)
					(
						block_type,
						entity,
						effectid,
						effectindex,
						target,
						suboption
					) = sre.groups()
			else:
				if opcode == "ACTION_START":
					sre = tokens.ACTION_START_RE.match(data)
				else:
					sre = tokens.BLOCK_START_12051_RE.match(data)

				if sre is None:
					sre = tokens.ACTION_START_OLD_RE.match(data)
					if not sre:
						raise RegexParsingError(data)
					entity, block_type, index, target = sre.groups()
				else:
					block_type, entity, effectid, effectindex, target = sre.groups()

			self.block_start(
				ps, ts, entity, block_type, index, effectid, effectindex, target, suboption,
				trigger_keyword
			)
			return
		elif opcode in ("ACTION_END", "BLOCK_END"):
			regex, callback = tokens.BLOCK_END_RE, self.block_end
		elif opcode == "FULL_ENTITY":
			if data.startswith("FULL_ENTITY - Updating"):
				regex, callback = tokens.FULL_ENTITY_UPDATE_RE, self.full_entity_update
			else:
				regex, callback = tokens.FULL_ENTITY_CREATE_RE, self.full_entity
		elif opcode == "SHOW_ENTITY":
			regex, callback = tokens.SHOW_ENTITY_RE, self.show_entity
		elif opcode == "HIDE_ENTITY":
			regex, callback = tokens.HIDE_ENTITY_RE, self.hide_entity
		elif opcode == "CHANGE_ENTITY":
			regex, callback = tokens.CHANGE_ENTITY_RE, self.change_entity
		elif opcode == "TAG_CHANGE":
			regex, callback = tokens.TAG_CHANGE_RE, self.tag_change
		elif opcode == "META_DATA":
			regex, callback = tokens.META_DATA_RE, self.meta_data
		elif opcode == "RESET_GAME":
			regex, callback = tokens.RESET_GAME_RE, self.reset_game
		elif opcode == "SUB_SPELL_START":
			regex, callback = tokens.SUB_SPELL_START_RE, self.sub_spell_start
		elif opcode == "SUB_SPELL_END":
			regex, callback = tokens.SUB_SPELL_END_RE, self.sub_spell_end
		elif opcode == "CACHED_TAG_FOR_DORMANT_CHANGE":
			regex, callback = (
				tokens.CACHED_TAG_FOR_DORMANT_CHANGE_RE,
				self.cached_tag_for_dormant_change
			)
		elif opcode == "VO_SPELL":
			regex, callback = tokens.VO_SPELL_RE, self.vo_spell
		elif opcode == "SHUFFLE_DECK":
			regex, callback = tokens.SHUFFLE_DECK_RE, self.shuffle_deck
		else:
			raise NotImplementedError(data)

		sre = regex.match(data)
		if not sre:
			logging.warning("[%s] Could not correctly parse %r", ts, data)
			return
		return callback(ps, ts, *sre.groups())

	# Messages

	@staticmethod
	def create_game(ps: ParsingState, ts):
		pt = packets.PacketTree(ts)
		pt.spectator_mode = ps.spectator_mode

		ps.games.append(pt)
		ps.current_block = pt
		ps.packet_tree = pt

	@staticmethod
	def block_end(ps, ts):
		ps.block_end(ts)

	@staticmethod
	def block_start(
		ps: ParsingState, ts, entity, block_type, index, effectid, effectindex, target,
		suboption, trigger_keyword
	):
		entity_id = ps.parse_entity_or_player(entity)
		block_type = parse_enum(BlockType, block_type)

		if index is not None:
			index = int(index)

		target = ps.parse_entity_or_player(target)

		if suboption is not None:
			suboption = int(suboption)
		if trigger_keyword is not None:
			trigger_keyword = parse_enum(GameTag, trigger_keyword)

		block = packets.Block(
			ts, entity_id, block_type, index, effectid, effectindex, target, suboption,
			trigger_keyword
		)
		block.parent = ps.current_block
		ps.register_packet(block)
		ps.current_block = block
		return block

	def full_entity(self, ps: ParsingState, ts, entity_id: int, card_id: str):
		ps.entity_packet = packets.FullEntity(ts, entity_id, card_id)
		ps.register_packet(ps.entity_packet)

		if self._creating_game:
			# First packet after create game should always be a FULL_ENTITY
			self._creating_game = False

			# Check if we got an abnormal amount of players
			player_count = len(ps.game_packet.players)
			if player_count < 2:
				msg = "Expected at least 2 players before the first entity, got %r"
				raise ParsingError(msg % player_count)

		return ps.entity_packet

	def full_entity_update(self, ps: ParsingState, ts, entity, card_id):
		entity_id = ps.parse_entity_id(entity)
		return self.full_entity(ps, ts, entity_id, card_id)

	@staticmethod
	def show_entity(ps: ParsingState, ts, entity, card_id):
		entity_id = ps.parse_entity_id(entity)
		ps.entity_packet = packets.ShowEntity(ts, entity_id, card_id)
		ps.register_packet(ps.entity_packet)
		return ps.entity_packet

	@staticmethod
	def hide_entity(ps: ParsingState, ts, entity, tag, value):
		entity_id = ps.parse_entity_id(entity)
		tag, value = parse_tag(tag, value)

		if tag != GameTag.ZONE:
			raise ParsingError("HIDE_ENTITY got non-zone tag (%r)" % tag)

		packet = packets.HideEntity(ts, entity_id, value)
		ps.register_packet(packet)
		return packet

	@staticmethod
	def change_entity(ps: ParsingState, ts, entity, card_id):
		entity_id = ps.parse_entity_or_player(entity)
		ps.entity_packet = packets.ChangeEntity(ts, entity_id, card_id)
		ps.register_packet(ps.entity_packet)
		return ps.entity_packet

	@staticmethod
	def meta_data(ps: ParsingState, ts, meta, data, info_count):
		meta = parse_enum(MetaDataType, meta)

		if meta == MetaDataType.JOUST:
			data = ps.parse_entity_id(data)

		count = int(info_count)
		ps.metadata_node = MetaData(ts, meta, data, count)
		ps.register_packet(ps.metadata_node)
		return ps.metadata_node

	@staticmethod
	def _register_player_on_tag_change(
		ps: ParsingState,
		player: PlayerReference,
		tag: str,
		value: str
	):
		# Triggers on every TAG_CHANGE where the corresponding entity is a LazyPlayer.
		# Will attempt to return a new value instead

		if tag == GameTag.ENTITY_ID:

			# This is the simplest check. When a player entity is declared,
			# its ENTITY_ID is not available immediately (in pre-6.0).
			# If we get a matching ENTITY_ID, then we can use that to match it.

			return ps.manager.create_or_update_player(
				name=player.name,
				entity_id=int(value)
			)
		elif tag == GameTag.LAST_CARD_PLAYED:

			# This is a fallback to register_player_name_mulligan in case the mulligan
			# phase is not available in this game (spectator mode, reconnects).

			player_id = ps.manager.get_controller_by_entity_id(int(value))

			if player_id is None:
				raise ParsingError("Unknown entity ID on TAG_CHANGE: %r" % value)

			return ps.manager.create_or_update_player(name=player.name, player_id=player_id)
		else:
			return player

	def tag_change(self, ps: ParsingState, ts, e, tag, value, def_change):
		entity = ps.parse_entity_or_player(e)
		tag, value = parse_tag(tag, value)
		self._check_for_mulligan_hack(ps, ts, tag, value)

		if tag == GameTag.CONTROLLER:
			entity_id = coerce_to_entity_id(entity)
			ps.manager.register_controller(int(entity_id), int(value))

		elif tag == GameTag.FIRST_PLAYER:
			entity_id = coerce_to_entity_id(entity)
			ps.manager.notify_first_player(int(entity_id))

		if isinstance(entity, PlayerReference):
			entity_id = self._register_player_on_tag_change(ps, entity, tag, value)
		else:
			entity_id = entity

		has_change_def = def_change == tokens.DEF_CHANGE
		packet = packets.TagChange(ts, entity_id, tag, value, has_change_def)
		ps.register_packet(packet)
		return packet

	@staticmethod
	def reset_game(ps: ParsingState, ts):
		packet = packets.ResetGame(ts)
		ps.register_packet(packet)
		return packet

	@staticmethod
	def sub_spell_start(ps: ParsingState, ts, spell_prefab_guid, source, target_count):
		source = int(source)
		target_count = int(target_count)

		sub_spell = packets.SubSpell(ts, spell_prefab_guid, source, target_count)
		sub_spell.parent = ps.current_block
		ps.register_packet(sub_spell)
		ps.current_block = sub_spell
		return sub_spell

	@staticmethod
	def sub_spell_end(ps: ParsingState, ts):
		if not ps.current_block.parent:
			logging.warning("[%s] Orphaned SUB_SPELL_END detected", ts)
			return ps.current_block

		if isinstance(ps.current_block, SubSpell):
			ps.current_block.end()

		sub_spell = ps.current_block
		ps.current_block = ps.current_block.parent
		return sub_spell

	@staticmethod
	def cached_tag_for_dormant_change(ps: ParsingState, ts, e, tag, value):
		entity_id = ps.parse_entity_id(e)
		tag, value = parse_tag(tag, value)

		packet = packets.CachedTagForDormantChange(ts, entity_id, tag, value)
		ps.register_packet(packet)
		return packet

	@staticmethod
	def vo_spell(ps: ParsingState, ts, brguid, vospguid, blocking, delayms):
		packet = packets.VOSpell(ts, brguid, vospguid, blocking == "True", int(delayms))
		ps.register_packet(packet)
		return packet

	@staticmethod
	def shuffle_deck(ps: ParsingState, ts, player_id):
		packet = packets.ShuffleDeck(ts, int(player_id))
		ps.register_packet(packet)
		return packet


class OptionsHandler(HandlerBase):
	def __init__(self):
		super().__init__()

		self._option_packet = None
		self._options_packet = None
		self._suboption_packet = None

	def find_callback(self, method):
		if method == self.parse_method("SendOption"):
			return self.handle_send_option
		elif method == self.parse_method("DebugPrintOptions"):
			return self.handle_options

	def _parse_option_packet(self, ps: ParsingState, ts, data):
		if " errorParam=" in data:
			sre = tokens.OPTIONS_OPTION_ERROR_RE.match(data)
			if not sre:
				raise RegexParsingError(data)
			optype, entity_id, option_type, entity, error, error_param = sre.groups()
			error, error_param = clean_option_errors(error, error_param)
		else:
			sre = tokens.OPTIONS_OPTION_RE.match(data)
			if not sre:
				raise RegexParsingError(data)
			optype, entity_id, option_type, entity = sre.groups()
			error, error_param = None, None

		entity_id = int(entity_id)
		option_type = parse_enum(OptionType, option_type)
		if entity:
			entity = ps.parse_entity_or_player(entity)
		self._option_packet = packets.Option(
			ts,
			entity,
			entity_id,
			option_type,
			optype,
			error,
			error_param
		)
		if not self._options_packet:
			raise ParsingError("Option without a parent option group: %r" % data)

		ps.register_packet(self._option_packet, node=self._options_packet.options)
		self._suboption_packet = None

		return self._option_packet

	def _parse_suboption_packet(self, ps: ParsingState, ts, data):
		if " errorParam=" in data:
			sre = tokens.OPTIONS_SUBOPTION_ERROR_RE.match(data)
			if not sre:
				raise RegexParsingError(data)
			optype, entity_id, entity, error, error_param = sre.groups()
			error, error_param = clean_option_errors(error, error_param)
		else:
			sre = tokens.OPTIONS_SUBOPTION_RE.match(data)
			if not sre:
				raise RegexParsingError(data)
			optype, entity_id, entity = sre.groups()
			error, error_param = None, None

		entity_id = int(entity_id)
		if entity:
			entity = ps.parse_entity_or_player(entity)

		node = None
		packet = packets.Option(ts, entity, entity_id, None, optype, error, error_param)

		if optype == "subOption":
			self._suboption_packet = packet
			node = self._option_packet
		elif optype == "target":
			node = self._suboption_packet or self._option_packet

		if not node:
			raise ParsingError("SubOption / target without a matching option: %r" % data)

		ps.register_packet(packet, node=node.options)

		return packet

	@staticmethod
	def _check_for_options_hack(ps: ParsingState, ts):

		# Battlegrounds games tend to omit the BLOCK_END just before options start. As
		# options will always be on the top level, we can safely close any remaining block
		# that is open at this time.

		if isinstance(ps.current_block, packets.Block):
			logging.warning("[%s] Broken option nesting. Working around...", ts)
			ps.block_end(ts)
			assert not isinstance(ps.current_block, packets.Block)

	def handle_options(self, ps: ParsingState, ts, data):
		self._check_for_options_hack(ps, ts)
		if data.startswith("id="):
			sre = tokens.OPTIONS_ENTITY_RE.match(data)
			if not sre:
				raise RegexParsingError(data)
			entity_id, = sre.groups()
			entity_id = int(entity_id)
			self._options_packet = packets.Options(ts, entity_id)
			ps.current_block.packets.append(self._options_packet)
		elif data.startswith("option "):
			return self._parse_option_packet(ps, ts, data)
		elif data.startswith(("subOption ", "target ")):
			return self._parse_suboption_packet(ps, ts, data)

	def handle_send_option(self, ps: ParsingState, ts, data):
		if data.startswith("selectedOption="):
			sre = tokens.SEND_OPTION_RE.match(data)
			if not sre:
				raise RegexParsingError(data)
			option, suboption, target, position = sre.groups()
			packet = packets.SendOption(
				ts,
				int(option),
				int(suboption),
				int(target),
				int(position)
			)
			ps.register_packet(packet)
			return packet
		raise NotImplementedError("Unhandled send option: %r" % data)


class ChoicesHandler(HandlerBase):
	def __init__(self):
		super().__init__()

	def find_callback(self, method):
		if method == self.parse_method("DebugPrintEntityChoices"):
			return self.handle_entity_choices
		elif method == self.parse_method("DebugPrintChoices"):
			return self.handle_entity_choices_old
		elif method == self.parse_method("SendChoices"):
			return self.handle_send_choices
		elif method == self.parse_method("DebugPrintEntitiesChosen"):
			return self.handle_entities_chosen

	def handle_entity_choices_old(self, ps: ParsingState, ts, data):
		if data.startswith("id="):
			sre = tokens.CHOICES_CHOICE_OLD_1_RE.match(data)
			if sre:
				self.register_choices_old_1(ps, ts, *sre.groups())
			else:
				sre = tokens.CHOICES_CHOICE_OLD_2_RE.match(data)
				if not sre:
					raise RegexParsingError(data)
				self.register_choices_old_2(ps, ts, *sre.groups())
		else:
			return self.handle_entity_choices(ps, ts, data)

	def handle_entity_choices(self, ps: ParsingState, ts, data):
		if data.startswith("id="):
			sre = tokens.CHOICES_CHOICE_RE.match(data)
			if not sre:
				raise RegexParsingError(data)
			return self.register_choices(ps, ts, *sre.groups())
		elif data.startswith("Source="):
			sre = tokens.CHOICES_SOURCE_RE.match(data)
			if not sre:
				raise RegexParsingError(data)
			entity, = sre.groups()
			entity_id = ps.parse_entity_or_player(entity)
			if not ps.choice_packet:
				raise ParsingError("Source Choice Entity outside of choie packet: %r" % data)
			ps.choice_packet.source = entity_id
			return entity_id
		elif data.startswith("Entities["):
			sre = tokens.CHOICES_ENTITIES_RE.match(data)
			if not sre:
				raise RegexParsingError(data)
			idx, entity = sre.groups()
			entity_id = ps.parse_entity_or_player(entity)
			if not entity_id:
				raise ParsingError("Missing choice entity %r (%r)" % (entity_id, entity))
			if not ps.choice_packet:
				raise ParsingError("Choice Entity outside of choice packet: %r" % data)
			ps.choice_packet.choices.append(entity_id)
			return entity_id
		raise NotImplementedError("Unhandled entity choice: %r" % data)

	@staticmethod
	def _register_choices(
		ps: ParsingState,
		ts,
		choice_id,
		player,
		tasklist,
		choice_type,
		min_choice,
		max_choice
	):
		choice_type = parse_enum(ChoiceType, choice_type)
		min_choice, max_choice = int(min_choice), int(max_choice)
		ps.choice_packet = packets.Choices(
			ts,
			player,
			choice_id,
			tasklist,
			choice_type,
			min_choice,
			max_choice
		)
		ps.register_packet(ps.choice_packet)
		return ps.choice_packet

	def register_choices_old_1(self, ps: ParsingState, ts, choice_id, choice_type):
		player = None

		# XXX: We don't have a player here for old games.
		# Is it safe to assume CURRENT_PLAYER?

		tasklist = None
		min_choice, max_choice = 0, 0

		return self._register_choices(
			ps,
			ts,
			int(choice_id),
			player,
			tasklist,
			choice_type,
			min_choice,
			max_choice
		)

	def register_choices_old_2(
		self,
		ps: ParsingState,
		ts,
		choice_id,
		player_id,
		choice_type,
		min_choice,
		max_choice
	):
		return self._register_choices(
			ps,
			ts,
			int(choice_id),
			ps.manager.get_player_by_player_id(int(player_id)),
			None,
			choice_type,
			min_choice,
			max_choice
		)

	def register_choices(
		self,
		ps: ParsingState,
		ts,
		choice_id,
		player,
		tasklist,
		choice_type,
		min_choice,
		max_choice
	):
		player = ps.parse_entity_or_player(player)

		if tasklist is not None:
			# Sometimes tasklist is empty

			tasklist = int(tasklist)

		return self._register_choices(
			ps,
			ts,
			int(choice_id),
			player,
			tasklist,
			choice_type,
			min_choice,
			max_choice
		)

	def handle_send_choices(self, ps: ParsingState, ts, data):
		if data.startswith("id="):
			sre = tokens.SEND_CHOICES_CHOICE_RE.match(data)

			if not sre:
				raise RegexParsingError(data)

			choice_id, choice_type = sre.groups()
			ps.send_choice_packet = packets.SendChoices(
				ts,
				int(choice_id),
				parse_enum(ChoiceType, choice_type)
			)
			ps.register_packet(ps.send_choice_packet)
			return ps.send_choice_packet
		elif data.startswith("m_chosenEntities"):
			sre = tokens.SEND_CHOICES_ENTITIES_RE.match(data)

			if not sre:
				raise RegexParsingError(data)

			idx, entity = sre.groups()
			ep = ps.parse_entity_or_player(entity)

			if not ep:
				raise ParsingError("Missing chosen entity %r (%r)" % (ep, entity))
			if not ps.send_choice_packet:
				raise ParsingError("Chosen Entity outside of choice packet: %r" % data)

			ps.send_choice_packet.choices.append(ep)
			return ep
		raise NotImplementedError("Unhandled send choice: %r" % data)

	def handle_entities_chosen(self, ps: ParsingState, ts, data):
		if data.startswith("id="):
			sre = tokens.ENTITIES_CHOSEN_RE.match(data)
			if not sre:
				raise RegexParsingError(data)
			choice_id, player, count = sre.groups()
			player = ps.parse_entity_or_player(player)

			ps.chosen_packet_count = int(count)
			ps.chosen_packet = packets.ChosenEntities(ts, player, int(choice_id))
			ps.register_packet(ps.chosen_packet)

			for player_id, mulligan_choice in ps.mulligan_choices.items():
				if mulligan_choice.id == ps.chosen_packet.id:
					ps.manager.create_or_update_player(
						name=player.name,
						player_id=player_id
					)
					break

			return ps.chosen_packet
		elif data.startswith("Entities["):
			sre = tokens.ENTITIES_CHOSEN_ENTITIES_RE.match(data)
			if not sre:
				raise RegexParsingError(data)
			idx, entity = sre.groups()
			entity_id = ps.parse_entity_or_player(entity)

			if not entity_id:
				raise ParsingError("Missing entity chosen %r (%r)" % (entity_id, entity))
			if not ps.chosen_packet:
				raise ParsingError("Entity Chosen outside of choice packet: %r" % data)

			ps.chosen_packet.choices.append(entity_id)

			if len(ps.chosen_packet.choices) > ps.chosen_packet_count:
				raise ParsingError(
					"Too many choices (expected %r)" % ps.chosen_packet_count
				)
			return entity_id
		raise NotImplementedError("Unhandled entities chosen: %r" % data)


class SpectatorModeHandler:

	def process_spectator_mode(self, ps: ParsingState, line):
		if line == tokens.SPECTATOR_MODE_BEGIN_GAME:
			ps.set_spectating(True)
		elif line == tokens.SPECTATOR_MODE_BEGIN_FIRST:
			ps.set_spectating(True, False)
		elif line == tokens.SPECTATOR_MODE_BEGIN_SECOND:
			ps.set_spectating(True, True)
		elif line == tokens.SPECTATOR_MODE_END_MODE:
			ps.set_spectating(False, False)
		elif line == tokens.SPECTATOR_MODE_END_GAME:
			ps.set_spectating(False, False)
		else:
			raise NotImplementedError("Unhandled spectator mode: %r" % line)


class LogParser:
	def __init__(self):
		self.line_regex = tokens.POWERLOG_LINE_RE
		self._current_date = None
		self._synced_timestamp = False
		self._last_ts = None

		self._parsing_state = ParsingState()

		self._power_handler = PowerHandler()
		self._choices_handler = ChoicesHandler()
		self._options_handler = OptionsHandler()
		self._spectator_mode_handler = SpectatorModeHandler()

	def flush(self):
		self._parsing_state.flush()

	@property
	def game_meta(self):
		return self._parsing_state.game_meta

	@property
	def games(self):
		return self._parsing_state.games

	def handle_entity_choices(self, ts, data):
		return self._choices_handler.handle_entity_choices(self._parsing_state, ts, data)

	def handle_options(self, ts, data):
		return self._options_handler.handle_options(self._parsing_state, ts, data)

	def handle_power(self, ts, opcode, data):
		return self._power_handler.handle_power(self._parsing_state, ts, opcode, data)

	def parse_entity_id(self, entity):
		return self._parsing_state.parse_entity_id(entity)

	def parse_timestamp(self, ts, _method):
		if self._last_ts is not None and self._last_ts[0] == ts:
			return self._last_ts[1]

		ret = parse_time(ts)

		if not self._synced_timestamp:

			# The first timestamp we parse requires syncing the time
			# (in case _current_date is greater than the start date)

			if self._current_date is not None:
				self._current_date = self._current_date.replace(
					hour=ret.hour,
					minute=ret.minute,
					second=ret.second,
					microsecond=ret.microsecond,
				)

			# Only do it once per parse tree

			self._synced_timestamp = True

		# Logs don't have dates :(

		if self._current_date is None:
			self._last_ts = (ts, ret)

			# No starting date is available. Return just the time.

			return ret

		ret = datetime.combine(self._current_date, ret)
		ret = ret.replace(tzinfo=self._current_date.tzinfo)

		if ret < self._current_date:

			# If the new date falls before the last saved date, that
			# means we rolled over and need to increment the day by 1.

			ret += timedelta(days=1)

		self._current_date = ret
		self._last_ts = (ts, ret)
		return ret

	@property
	def player_manager(self):
		return self._parsing_state.manager

	def read(self, fp):
		for line in fp:
			self.read_line(line)

	def read_line(self, line):
		sre = tokens.TIMESTAMP_RE.match(line)

		if not sre:
			raise RegexParsingError(line)

		level, ts, line = sre.groups()

		if line.startswith(tokens.SPECTATOR_MODE_TOKEN):
			line = line.replace(tokens.SPECTATOR_MODE_TOKEN, "").strip()
			return self._spectator_mode_handler.process_spectator_mode(
				self._parsing_state,
				line
			)

		sre = self.line_regex.match(line)

		if not sre:
			return

		method, msg = sre.groups()
		msg = msg.strip()

		if not self._parsing_state.current_block and "CREATE_GAME" not in msg:

			# Ignore messages before the first CREATE_GAME packet

			return

		for handler in self._power_handler, self._choices_handler, self._options_handler:
			callback = handler.find_callback(method)
			if callback:
				ts = self.parse_timestamp(ts, method)
				return callback(self._parsing_state, ts, msg)



from typing import Optional

from hearthstone.entities import Card, Game, Player
from hearthstone.enums import BlockType, GameTag, Zone

from . import packets
from .exceptions import ExporterError, MissingPlayerData
from .player import PlayerManager, coerce_to_entity_id


class BaseExporter:
	def __init__(self, packet_tree):
		self.packet_tree = packet_tree
		self.dispatch = self.get_dispatch_dict()

	def get_dispatch_dict(self):
		return {
			packets.CreateGame: self.handle_create_game,
			packets.CreateGame.Player: self.handle_player,
			packets.Block: self.handle_block,
			packets.FullEntity: self.handle_full_entity,
			packets.HideEntity: self.handle_hide_entity,
			packets.ShowEntity: self.handle_show_entity,
			packets.ChangeEntity: self.handle_change_entity,
			packets.TagChange: self.handle_tag_change,
			packets.MetaData: self.handle_metadata,
			packets.Choices: self.handle_choices,
			packets.SendChoices: self.handle_send_choices,
			packets.ChosenEntities: self.handle_chosen_entities,
			packets.Options: self.handle_options,
			packets.Option: self.handle_option,
			packets.SendOption: self.handle_send_option,
			packets.ResetGame: self.handle_reset_game,
			packets.SubSpell: self.handle_sub_spell,
			packets.CachedTagForDormantChange: self.handle_cached_tag_for_dormant_change,
			packets.VOSpell: self.handle_vo_spell,
			packets.ShuffleDeck: self.handle_shuffle_deck
		}

	def export(self):
		for packet in self.packet_tree:
			self.export_packet(packet)
		self.flush()
		return self

	def export_packet(self, packet):
		packet_type = packet.__class__
		handler = self.dispatch.get(packet_type, None)
		if not handler:
			raise NotImplementedError("Don't know how to export %r" % packet_type)
		handler(packet)

	def flush(self):
		"""Finalize the export and allow any intermediate state to be cleaned up."""
		pass

	def handle_create_game(self, packet):
		pass

	def handle_player(self, packet):
		pass

	def handle_block(self, packet):
		for p in packet.packets:
			self.export_packet(p)

	def handle_full_entity(self, packet):
		pass

	def handle_hide_entity(self, packet):
		pass

	def handle_show_entity(self, packet):
		pass

	def handle_change_entity(self, packet):
		pass

	def handle_tag_change(self, packet):
		pass

	def handle_metadata(self, packet):
		pass

	def handle_choices(self, packet):
		pass

	def handle_send_choices(self, packet):
		pass

	def handle_chosen_entities(self, packet):
		pass

	def handle_options(self, packet):
		pass

	def handle_option(self, packet):
		pass

	def handle_send_option(self, packet):
		pass

	def handle_reset_game(self, packet):
		pass

	def handle_sub_spell(self, packet):
		for p in packet.packets:
			self.export_packet(p)

	def handle_cached_tag_for_dormant_change(self, packet):
		pass

	def handle_vo_spell(self, packet):
		pass

	def handle_shuffle_deck(self, packet):
		pass


class CompositeExporter(BaseExporter):
	"""Exporter implementation that broadcasts packets to configured child exporters

	Use this class to compose multiple exporters in order to do a single pass over a
	configured packet tree. Note:

	- Packet trees passed to constructors of child exporters will be ignored; only the
	packets in the packet tree passed to this class's constructor will be visited.
	- Unlike BaseExporter, the `handle_block` and `handle_sub_spell` methods on this class
	do not recursively invoke the `handle_packet` on child packets; child exporters used in
	an instance of CompositeExporter *should* continue to recursively invoke `handle_packet`
	however (or delegate to their superclass implementation).
	"""

	def __init__(self, packet_tree, exporters):
		super().__init__(packet_tree)
		self.exporters = exporters

	def flush(self):
		for exporter in self.exporters:
			exporter.flush()

	def handle_create_game(self, packet):
		for exporter in self.exporters:
			exporter.handle_create_game(packet)

	def handle_player(self, packet):
		for exporter in self.exporters:
			exporter.handle_player(packet)

	def handle_block(self, packet):
		for exporter in self.exporters:
			exporter.handle_block(packet)

	def handle_full_entity(self, packet):
		for exporter in self.exporters:
			exporter.handle_full_entity(packet)

	def handle_hide_entity(self, packet):
		for exporter in self.exporters:
			exporter.handle_hide_entity(packet)

	def handle_show_entity(self, packet):
		for exporter in self.exporters:
			exporter.handle_show_entity(packet)

	def handle_change_entity(self, packet):
		for exporter in self.exporters:
			exporter.handle_change_entity(packet)

	def handle_tag_change(self, packet):
		for exporter in self.exporters:
			exporter.handle_tag_change(packet)

	def handle_metadata(self, packet):
		for exporter in self.exporters:
			exporter.handle_metadata(packet)

	def handle_choices(self, packet):
		for exporter in self.exporters:
			exporter.handle_choices(packet)

	def handle_send_choices(self, packet):
		for exporter in self.exporters:
			exporter.handle_send_choices(packet)

	def handle_chosen_entities(self, packet):
		for exporter in self.exporters:
			exporter.handle_chosen_entities(packet)

	def handle_options(self, packet):
		for exporter in self.exporters:
			exporter.handle_options(packet)

	def handle_option(self, packet):
		for exporter in self.exporters:
			exporter.handle_option(packet)

	def handle_send_option(self, packet):
		for exporter in self.exporters:
			exporter.handle_send_option(packet)

	def handle_reset_game(self, packet):
		for exporter in self.exporters:
			exporter.handle_reset_game(packet)

	def handle_sub_spell(self, packet):
		for exporter in self.exporters:
			exporter.handle_sub_spell(packet)

	def handle_cached_tag_for_dormant_change(self, packet):
		for exporter in self.exporters:
			exporter.handle_cached_tag_for_dormant_change(packet)

	def handle_vo_spell(self, packet):
		for exporter in self.exporters:
			exporter.handle_vo_spell(packet)

	def handle_shuffle_deck(self, packet):
		for exporter in self.exporters:
			exporter.handle_shuffle_deck(packet)


class EntityTreeExporter(BaseExporter):
	game_class = Game
	player_class = Player
	card_class = Card

	class EntityNotFound(Exception):
		pass

	def __init__(self, packet_tree, player_manager: Optional[PlayerManager] = None):
		super().__init__(packet_tree)

		self.game: Optional[Game] = None

		self.player_manager = player_manager

	def find_entity(self, entity_id: int, opcode):
		try:
			entity = self.game.find_entity_by_id(entity_id)
		except MissingPlayerData:
			raise self.EntityNotFound("Error getting entity %r for %s" % (id, opcode))
		if not entity:
			raise self.EntityNotFound("Attempting %s on entity %r (not found)" % (opcode, id))
		return entity

	def handle_block(self, packet):
		if packet.type == BlockType.GAME_RESET:
			self.game.reset()
		super().handle_block(packet)

	def handle_create_game(self, packet):
		self.game = self.game_class(packet.entity)
		self.game.create(packet.tags)
		for player in packet.players:
			self.export_packet(player)
		return self.game

	def handle_player(self, packet):
		entity_id = coerce_to_entity_id(packet.entity)

		# If we have a PlayerManager, first we mutate the CreateGame.Player packet.
		# This will have to change if we're ever able to immediately get the names.

		if self.player_manager:
			player = self.player_manager.get_player_by_entity_id(int(entity_id))
			packet.name = player.name

		entity = self.player_class(
			entity_id,
			packet.player_id,
			packet.hi,
			packet.lo,
			packet.name
		)
		entity.tags = dict(packet.tags)
		self.game.register_entity(entity)
		entity.initial_hero_entity_id = entity.tags.get(GameTag.HERO_ENTITY, 0)
		return entity

	def handle_full_entity(self, packet):
		entity_id = packet.entity

		# Check if the entity already exists in the game first.
		# This prevents creating it twice.
		# This can legitimately happen in case of GAME_RESET
		existing_entity = self.game.find_entity_by_id(entity_id)
		if existing_entity is not None:
			existing_entity.card_id = packet.card_id
			existing_entity.tags = dict(packet.tags)
			return existing_entity

		entity = self.card_class(int(entity_id), packet.card_id)
		entity.tags = dict(packet.tags)
		self.game.register_entity(entity)
		return entity

	def handle_hide_entity(self, packet):
		entity = self.find_entity(packet.entity, "HIDE_ENTITY")
		entity.hide()
		return entity

	def handle_show_entity(self, packet):
		entity = self.find_entity(packet.entity, "SHOW_ENTITY")
		entity.reveal(packet.card_id, dict(packet.tags))
		return entity

	def handle_change_entity(self, packet):
		entity = self.find_entity(packet.entity, "CHANGE_ENTITY")
		if not entity.card_id:
			# This can only happen if the entity's initial reveal was missed.
			# Not raising this can cause entities to have a card id, but no initial_card_id.
			# If you want that behaviour, subclass EntityTreeExporter and override this.
			raise ExporterError(
				f"CHANGE_ENTITY {packet.entity} to {packet.card_id} with no previous known CardID."
			)
		entity.change(packet.card_id, dict(packet.tags))
		return entity

	def handle_tag_change(self, packet):
		entity_id = coerce_to_entity_id(packet.entity)
		entity = self.find_entity(int(entity_id), "TAG_CHANGE")
		entity.tag_change(packet.tag, packet.value)

		return entity


class FriendlyPlayerExporter(BaseExporter):
	"""
	An exporter that will attempt to guess the friendly player in the game by
	looking for initial unrevealed cards.
	May produce incorrect results in spectator mode if both hands are revealed.
	"""
	def __init__(self, packet_tree):
		super().__init__(packet_tree)
		self._controller_map = {}
		self.friendly_player = None
		self._ai_player = None
		self._non_ai_players = []

	def export(self):
		for packet in self.packet_tree:
			self.export_packet(packet)
			if self.friendly_player is not None:
				# Stop export once we have it
				break
		return self.friendly_player

	def handle_create_game(self, packet):
		for player in packet.players:
			self.export_packet(player)
		if self._ai_player is not None and len(self._non_ai_players) == 1:
			self.friendly_player = self._non_ai_players[0]

	def handle_player(self, packet):
		if packet.lo == 0:
			self._ai_player = packet.player_id
		else:
			self._non_ai_players.append(packet.player_id)

	def handle_tag_change(self, packet):
		if packet.tag == GameTag.CONTROLLER:
			self._controller_map[packet.entity] = packet.value

	def handle_full_entity(self, packet):
		tags = dict(packet.tags)
		if GameTag.CONTROLLER in tags:
			self._controller_map[packet.entity] = tags[GameTag.CONTROLLER]

	def handle_show_entity(self, packet):
		tags = dict(packet.tags)
		if GameTag.CONTROLLER in tags:
			self._controller_map[packet.entity] = tags[GameTag.CONTROLLER]

		if tags.get(GameTag.ZONE) != Zone.HAND:
			# Ignore cards already in play (such as enchantments, common in TB)
			return

		# The first SHOW_ENTITY packet will always be the friendly player's.
		self.friendly_player = self._controller_map[packet.entity]



from hearthstone.enums import PowerType


class PacketTree:
	def __init__(self, ts):
		self.ts = ts
		self.packets = []
		self.parent = None
		self.packet_counter = 0

	def __iter__(self):
		for packet in self.packets:
			yield packet

	@property
	def start_time(self):
		for packet in self.packets:
			if packet.ts:
				return packet.ts

	@property
	def end_time(self):
		for packet in self.packets[::-1]:
			if packet.ts:
				return packet.ts

	def export(self, cls=None):
		if cls is None:
			from .export import EntityTreeExporter as cls
		exporter = cls(self)
		return exporter.export()

	def recursive_iter(self, cls=None):
		"""
		Iterate recursively over the PacketTree.
		Optional argument `cls` yields only packets of that instance.
		"""
		if cls is None:
			cls = Packet

		def _iter_recursive(packets, cls):
			for packet in packets:
				if isinstance(packet, cls):
					yield packet
				if isinstance(packet, Block):
					for p in _iter_recursive(packet.packets, cls):
						yield p

		for packet in _iter_recursive(self.packets, cls):
			yield packet


class Packet:
	power_type = 0

	def __repr__(self):
		return "<%s>" % self.__class__.__name__


class Block(Packet):
	power_type = PowerType.BLOCK_START

	def __init__(
		self, ts, entity, type, index, effectid, effectindex, target, suboption, trigger_keyword
	):
		self.ts = ts
		self.entity = entity
		self.type = type
		self.index = index
		self.effectid = effectid
		self.effectindex = effectindex
		self.target = target
		self.suboption = suboption
		self.trigger_keyword = trigger_keyword
		self.ended = False
		self.packets = []

	def __iter__(self):
		for packet in self.packets:
			yield packet

	def __repr__(self):
		return "%s(entity=%r, type=%r, index=%r, target=%r)" % (
			self.__class__.__name__, self.entity, self.type, self.index, self.target
		)

	def end(self):
		self.ended = True

	def _export(self, game):
		for packet in self.packets:
			packet._export(game)


class MetaData(Packet):
	power_type = PowerType.META_DATA

	def __init__(self, ts, meta, data, count):
		self.ts = ts
		self.meta = meta
		self.data = data
		self.count = count
		self.info = []

	def __repr__(self):
		return "%s(meta=%r, data=%r)" % (self.__class__.__name__, self.meta, self.data)


class CreateGame(Packet):
	power_type = PowerType.CREATE_GAME

	class Player(Packet):
		def __init__(self, ts, entity, player_id, hi, lo):
			self.ts = ts
			self.entity = entity
			self.player_id = player_id
			self.hi = hi
			self.lo = lo
			self.tags = []
			self.name = None

	def __init__(self, ts, entity):
		self.ts = ts
		self.entity = entity
		self.tags = []
		self.players = []


class HideEntity(Packet):
	power_type = PowerType.HIDE_ENTITY

	def __init__(self, ts, entity, zone):
		self.ts = ts
		self.entity = entity
		self.zone = zone


class FullEntity(Packet):
	power_type = PowerType.FULL_ENTITY

	def __init__(self, ts, entity, card_id):
		self.ts = ts
		self.entity = entity
		self.card_id = card_id
		self.tags = []


class ShowEntity(Packet):
	power_type = PowerType.SHOW_ENTITY

	def __init__(self, ts, entity, card_id):
		self.ts = ts
		self.entity = entity
		self.card_id = card_id
		self.tags = []

	def __repr__(self):
		return "%s(entity=%r, card_id=%r)" % (
			self.__class__.__name__, self.entity, self.card_id
		)


class ChangeEntity(Packet):
	power_type = PowerType.CHANGE_ENTITY

	def __init__(self, ts, entity, card_id):
		self.ts = ts
		self.entity = entity
		self.card_id = card_id
		self.tags = []


class TagChange(Packet):
	power_type = PowerType.TAG_CHANGE

	def __init__(self, ts, entity, tag, value, has_change_def=False):
		self.ts = ts
		self.entity = entity
		self.tag = tag
		self.value = value
		self.has_change_def = has_change_def


class Choices(Packet):
	def __init__(self, ts, entity, id, tasklist, type, min, max):
		self.ts = ts
		self.entity = entity
		self.id = id
		self.tasklist = tasklist
		self.type = type
		self.min = min
		self.max = max
		self.source = None
		self.choices = []

	@property
	def player(self):
		return self.entity


class SendChoices(Packet):
	def __init__(self, ts, id, type):
		self.ts = ts
		self.entity = None
		self.id = id
		self.type = type
		self.choices = []


class ChosenEntities(Packet):
	def __init__(self, ts, entity, id):
		self.ts = ts
		self.entity = entity
		self.id = id
		self.choices = []


class Options(Packet):
	def __init__(self, ts, id):
		self.ts = ts
		self.entity = None
		self.id = id
		self.options = []


class Option(Packet):
	def __init__(self, ts, entity, id, type, optype, error, error_param):
		self.ts = ts
		self.entity = entity
		self.id = id
		self.type = type
		self.optype = optype
		self.error = error
		self.error_param = error_param
		self.options = []


class SendOption(Packet):
	def __init__(self, ts, option, suboption, target, position):
		self.ts = ts
		self.entity = None
		self.option = option
		self.suboption = suboption
		self.target = target
		self.position = position


class ResetGame(Packet):
	def __init__(self, ts):
		self.ts = ts


class SubSpell(Packet):
	power_type = PowerType.SUB_SPELL_START

	def __init__(self, ts, spell_prefab_guid, source, target_count):
		self.ts = ts
		self.spell_prefab_guid = spell_prefab_guid
		self.source = source
		self.target_count = target_count
		self.ended = False
		self.targets = []
		self.packets = []

	def __repr__(self):
		return "%s(spell_prefab_guid=%r, source=%r)" % (
			self.__class__.__name__, self.spell_prefab_guid, self.source
		)

	def end(self):
		self.ended = True


class CachedTagForDormantChange(Packet):
	power_type = PowerType.CACHED_TAG_FOR_DORMANT_CHANGE

	def __init__(self, ts, entity, tag, value):
		self.ts = ts
		self.entity = entity
		self.tag = tag
		self.value = value


class VOSpell(Packet):
	power_type = PowerType.VO_SPELL

	def __init__(self, ts, brguid, vospguid, blocking, delayms):
		self.ts = ts
		self.brguid = brguid
		self.vospguid = vospguid
		self.blocking = blocking
		self.delayms = delayms


class ShuffleDeck(Packet):
	power_type = PowerType.SHUFFLE_DECK

	def __init__(self, ts, player_id):
		self.ts = ts
		self.player_id = player_id



from hearthstone.enums import TAG_TYPES, GameTag


def parse_enum(enum, value):
	if value.isdigit():
		value = int(value)
	elif hasattr(enum, value):
		value = getattr(enum, value)
	else:
		raise Exception("Unhandled %s: %r" % (enum, value))
	return value


def parse_tag(tag, value):
	tag = parse_enum(GameTag, tag)
	if tag in TAG_TYPES:
		value = parse_enum(TAG_TYPES[tag], value)
	elif value.isdigit():
		value = int(value)
	else:
		raise NotImplementedError("Invalid string value %r = %r" % (tag, value))
	return tag, value



import re


# Entity format
GAME_ENTITY = "GameEntity"
UNKNOWN_HUMAN_PLAYER = "UNKNOWN HUMAN PLAYER"
_E = r"(%s|%s|\[.+\]|\d+|.+)" % (GAME_ENTITY, UNKNOWN_HUMAN_PLAYER)
ENTITY_RE = re.compile(r"\[.*\s*id=(\d+)\s*.*\]")

# Line format
TIMESTAMP_POWERLOG_FORMAT = r"%H:%M:%S.%f"
TIMESTAMP_RE = re.compile(r"^([DWE]) ([\d:.]+) (.+)$")
POWERLOG_LINE_RE = re.compile(r"([^(]+)\(\) - (.+)$")

# Game / Player
GAME_ENTITY_RE = re.compile(r"GameEntity EntityID=(\d+)")
PLAYER_ENTITY_RE = re.compile(r"Player EntityID=(\d+) PlayerID=(\d+) GameAccountId=\[hi=(\d+) lo=(\d+)\]$")

# Messages
CREATE_GAME_RE = re.compile(r"^CREATE_GAME$")
ACTION_START_OLD_RE = re.compile(r"ACTION_START Entity=%s (?:SubType|BlockType)=(\w+) Index=(-1|\d+) Target=%s$" % (_E, _E))
ACTION_START_RE = re.compile(r"ACTION_START SubType=(\w+) Entity=%s EffectCardId=(.*) EffectIndex=(-1|\d+) Target=%s$" % (_E, _E))
BLOCK_START_12051_RE = re.compile(r"BLOCK_START BlockType=(\w+) Entity=%s EffectCardId=(.*) EffectIndex=(-1|\d+) Target=%s$" % (_E, _E))
BLOCK_START_20457_RE = re.compile(r"BLOCK_START BlockType=(\w+) Entity=%s EffectCardId=(.*) EffectIndex=(-1|\d+) Target=%s SubOption=(-1|\d+)$" % (_E, _E))
BLOCK_START_20457_TRIGGER_KEYWORD_RE = re.compile(r"BLOCK_START BlockType=(\w+) Entity=%s EffectCardId=(.*) EffectIndex=(-1|\d+) Target=%s SubOption=(-1|\d+) TriggerKeyword=(\w+)$" % (_E, _E))
BLOCK_END_RE = re.compile(r"^(?:ACTION|BLOCK)_END$")
FULL_ENTITY_CREATE_RE = re.compile(r"FULL_ENTITY - Creating ID=(\d+) CardID=(\w+)?$")
FULL_ENTITY_UPDATE_RE = re.compile(r"FULL_ENTITY - Updating %s CardID=(\w+)?$" % _E)
SHOW_ENTITY_RE = re.compile(r"SHOW_ENTITY - Updating Entity=%s CardID=(\w+)$" % _E)
HIDE_ENTITY_RE = re.compile(r"HIDE_ENTITY - Entity=%s tag=(\w+) value=(\w+)$" % _E)
CHANGE_ENTITY_RE = re.compile(r"CHANGE_ENTITY - Updating Entity=%s CardID=(\w+)$" % _E)
DEF_CHANGE = "DEF CHANGE"
TAG_CHANGE_RE = re.compile(r"TAG_CHANGE Entity=%s tag=(\w+) value=(\w+) ?(%s)?" % (_E, DEF_CHANGE))
META_DATA_RE = re.compile(r"META_DATA - Meta=(\w+) Data=%s InfoCount=(\d+)" % _E)
RESET_GAME_RE = re.compile(r"RESET_GAME$")
SUB_SPELL_START_RE = re.compile(r"SUB_SPELL_START - SpellPrefabGUID=([\w:.]+) Source=(\d+) TargetCount=(\d+)$")
SUB_SPELL_END_RE = re.compile(r"SUB_SPELL_END$")
CACHED_TAG_FOR_DORMANT_CHANGE_RE = re.compile(r"CACHED_TAG_FOR_DORMANT_CHANGE Entity=%s tag=(\w+) value=(\w+)" % _E)
VO_SPELL_RE = re.compile(r"VO_SPELL - BrassRingGuid=(.*) - VoSpellPrefabGUID=(\w*)? - Blocking=(True|False) - AdditionalDelayInMs=(\d+)$")
SHUFFLE_DECK_RE = re.compile(r"SHUFFLE_DECK PlayerID=(\d+)$")

# Message details
TAG_VALUE_RE = re.compile(r"tag=(\w+) value=(\w+)")
METADATA_INFO_RE = re.compile(r"Info\[(\d+)\] = %s" % _E)
SUB_SPELL_START_SOURCE_RE = re.compile(r"Source = %s" % _E)
SUB_SPELL_START_TARGETS_RE = re.compile(r"Targets\[(\d+)\] = %s" % _E)

# Game
GAME_PLAYER_META = re.compile(r"PlayerID=(\d+), PlayerName=(.*)")

# Choices
CHOICES_CHOICE_OLD_1_RE = re.compile(r"id=(\d+) ChoiceType=(\w+)$")
CHOICES_CHOICE_OLD_2_RE = re.compile(r"id=(\d+) PlayerId=(\d+) ChoiceType=(\w+) CountMin=(\d+) CountMax=(\d+)$")
CHOICES_CHOICE_RE = re.compile(r"id=(\d+) Player=%s TaskList=(\d+)? ChoiceType=(\w+) CountMin=(\d+) CountMax=(\d+)$" % _E)
CHOICES_SOURCE_RE = re.compile(r"Source=%s$" % _E)
CHOICES_ENTITIES_RE = re.compile(r"Entities\[(\d+)\]=(\[.+\])$")
SEND_CHOICES_CHOICE_RE = re.compile(r"id=(\d+) ChoiceType=(.+)$")
SEND_CHOICES_ENTITIES_RE = re.compile(r"m_chosenEntities\[(\d+)\]=(\[.+\])$")
ENTITIES_CHOSEN_RE = re.compile(r"id=(\d+) Player=%s EntitiesCount=(\d+)$" % _E)
ENTITIES_CHOSEN_ENTITIES_RE = re.compile(r"Entities\[(\d+)\]=%s$" % _E)

# Options
OPTIONS_ENTITY_RE = re.compile(r"id=(\d+)$")
OPTIONS_OPTION_RE = re.compile(r"(option) (\d+) type=(\w+) mainEntity=%s?$" % _E)
OPTIONS_OPTION_ERROR_RE = re.compile(r"(option) (\d+) type=(\w+) mainEntity=%s? error=(\w+) errorParam=(\d+)?$" % _E)
OPTIONS_SUBOPTION_RE = re.compile(r"(subOption|target) (\d+) entity=%s?$" % _E)
OPTIONS_SUBOPTION_ERROR_RE = re.compile(r"(subOption|target) (\d+) entity=%s? error=(\w+) errorParam=(\d+)?$" % _E)
SEND_OPTION_RE = re.compile(r"selectedOption=(\d+) selectedSubOption=(-1|\d+) selectedTarget=(\d+) selectedPosition=(\d+)")

# Spectator mode
SPECTATOR_MODE_TOKEN = "=================="
SPECTATOR_MODE_BEGIN_GAME = "Start Spectator Game"
SPECTATOR_MODE_BEGIN_FIRST = "Begin Spectating 1st player"
SPECTATOR_MODE_BEGIN_SECOND = "Begin Spectating 2nd player"
SPECTATOR_MODE_END_MODE = "End Spectator Mode"
SPECTATOR_MODE_END_GAME = "End Spectator Game"



"""
Classes to provide lazy players that are treatable as an entity ID but
do not have to receive one immediately.
"""
from typing import Dict, List, Optional, Union

from hearthstone.enums import GameType

from .exceptions import MissingPlayerData
from .tokens import UNKNOWN_HUMAN_PLAYER


class PlayerReference:
	def __init__(
		self,
		entity_id: Optional[int] = None,
		name: Optional[str] = None,
		player_id: Optional[int] = None
	):
		self.entity_id = entity_id
		self.name = name
		self.player_id = player_id

	def __eq__(self, other):
		if not isinstance(other, PlayerReference):
			return False

		return (
			self.entity_id == other.entity_id and
			self.name == other.name and
			self.player_id == other.player_id
		)

	def __hash__(self):
		return hash((self.entity_id, self.name, self.player_id))

	def __repr__(self):
		return "%s(name=%r, entity_id=%r, player_id=%r)" % (
			self.__class__.__name__,
			self.name,
			self.entity_id,
			self.player_id
		)


def coerce_to_entity_id(entity_id_or_player: Union[int, PlayerReference]) -> int:
	if isinstance(entity_id_or_player, PlayerReference):
		if entity_id_or_player.entity_id is None:
			raise MissingPlayerData(
				"Entity ID not available for player %r" % entity_id_or_player.name
			)

		return entity_id_or_player.entity_id
	else:
		return entity_id_or_player


class InconsistentEntityIdError(Exception):

	def __init__(self, player: PlayerReference, entity_id: int):
		self.player = player
		self.entity_id = entity_id


class InconsistentPlayerIdError(Exception):

	def __init__(self, player: PlayerReference, player_id: int):
		self.player = player
		self.player_id = player_id


class PlayerManager:
	def __init__(self):
		self._entity_controller_map: Dict[int, int] = {}
		self._game_type = None
		self._name_aliases: Dict[str, str] = {}
		self._players_by_name: Dict[str, PlayerReference] = {}
		self._players_by_entity_id: Dict[int, PlayerReference] = {}
		self._players_by_player_id: Dict[int, PlayerReference] = {}
		self._player_resolution_order: List[int] = []

		self.ai_player: Optional[PlayerReference] = None
		self.first_player: Optional[PlayerReference] = None

	def _maybe_alias_name(self, name):
		if "#" in name:
			alias = name[:name.index("#")]
			if alias in self._name_aliases:
				assert self._name_aliases[alias] == name
			else:
				self._name_aliases[alias] = name

	def get_player_by_entity_id(self, entity_id: int) -> Optional[PlayerReference]:
		return self._players_by_entity_id.get(entity_id)

	def get_player_by_player_id(self, player_id: int) -> Optional[PlayerReference]:
		return self._players_by_player_id.get(player_id)

	def _guess_player_entity_id(self, name: str) -> Optional[PlayerReference]:
		assert name, "Expected a name for get_player_by_name (got %r)" % name
		if name not in self._players_by_name:
			if (
				len(self._players_by_name) == 1 and
				name != UNKNOWN_HUMAN_PLAYER and
				self._game_type != GameType.GT_BATTLEGROUNDS
			):
				# Maybe we can figure the name out right there and then.

				# NOTE: This is a neat trick, but it doesn't really work (and leads to
				# errors) when there are lots of players such that we can't predict the
				# entity ids. Hence the check for Battlegrounds above.

				other_player = next(iter(self._players_by_name.values()))
				entity_id = 3 if other_player.entity_id == 2 else 2
				if entity_id in self._players_by_entity_id:
					player = self._players_by_entity_id[entity_id]
					player.name = name

					self._players_by_name[name] = player
					self._maybe_alias_name(name)

					return player
				else:
					return self.create_or_update_player(
						name=name,
						entity_id=entity_id
					)
			elif len(self._players_by_name) > 1 and self.ai_player:
				# If we are registering our 3rd (or more) name, and we are in an AI game...
				# then it's probably the innkeeper with a new name.

				return self.create_or_update_player(
					name=name,
					entity_id=self.ai_player.entity_id
				)
			else:
				return None

	def create_or_update_player(
		self,
		name: Optional[str] = None,
		entity_id: Optional[int] = None,
		player_id: Optional[int] = None,
		is_ai: bool = False
	) -> PlayerReference:
		assert name or entity_id or player_id

		player: Optional[PlayerReference] = None

		if name and name != UNKNOWN_HUMAN_PLAYER:
			if name in self._players_by_name:
				player = self._players_by_name[name]
			elif name in self._name_aliases:
				player = self._players_by_name[self._name_aliases[name]]
			else:
				player = None

		if player is None:
			if entity_id and entity_id in self._players_by_entity_id:
				player = self._players_by_entity_id[entity_id]
			elif player_id and player_id in self._players_by_player_id:
				player = self._players_by_player_id[player_id]
			else:
				player = PlayerReference(
					name=name,
					entity_id=entity_id,
					player_id=player_id,
				)

		if is_ai:
			self.ai_player = player

		if name:
			if player.name is None or player.name == UNKNOWN_HUMAN_PLAYER:
				player.name = name
			elif player.name != name and player.name != self._name_aliases.get(name):
				if player == self.ai_player:

					# Need to check whether this is just the Innkeeper's name changing,
					# which happens for all sorts of valid reasons

					player.name = name
				else:
					raise AssertionError()

			if (
				name != UNKNOWN_HUMAN_PLAYER and
				name not in self._players_by_name and
				name not in self._name_aliases
			):
				if not player.entity_id and not entity_id:

					# We don't have an entity_id but we do have a name; let's see if we
					# can guess the entity id...

					self._guess_player_entity_id(name)
					if name in self._players_by_name:
						player = self._players_by_name[name]

				self._players_by_name[name] = player
				self._maybe_alias_name(name)
			elif player.name != name and player.name != self._name_aliases.get(name):
				raise AssertionError()

		if entity_id:
			if player.entity_id is None:
				player.entity_id = entity_id
			elif player.entity_id != entity_id:
				raise InconsistentEntityIdError(player, entity_id)

			if player.entity_id in self._players_by_entity_id:
				self._safe_merge_player_references(
					self._players_by_entity_id[player.entity_id],
					player,
				)
			else:
				self._players_by_entity_id[entity_id] = player

		if player_id:
			if player.player_id is None:
				player.player_id = player_id
			elif player.player_id != player_id:
				raise InconsistentPlayerIdError(player, player_id)

			if player.player_id in self._players_by_player_id:
				self._safe_merge_player_references(
					self._players_by_player_id[player.player_id],
					player,
				)
			else:
				self._players_by_player_id[player_id] = player

		if (
			player.name is not None and
			player.name != UNKNOWN_HUMAN_PLAYER and
			player.entity_id and
			player.player_id and
			player.player_id not in self._player_resolution_order
		):
			self._player_resolution_order.append(player.player_id)
		elif (
			player.name == UNKNOWN_HUMAN_PLAYER and
			player.entity_id is None and
			player.player_id is None and
			self._game_type != GameType.GT_BATTLEGROUNDS and
			len(self._player_resolution_order) < len(self._players_by_player_id)
		):
			unresolved_keys = [
				k for k in self._players_by_player_id.keys()
				if k not in self._player_resolution_order
			]

			if len(unresolved_keys) == 1:
				player = self._players_by_player_id[unresolved_keys[0]]

		return player

	@staticmethod
	def _safe_merge_player_references(left: PlayerReference, right: PlayerReference):
		if left.entity_id is None:
			if right.entity_id is not None:
				left.entity_id = right.entity_id
		elif right.entity_id is None:
			right.entity_id = left.entity_id
		else:
			assert left.entity_id == right.entity_id

		if left.player_id is None:
			if right.player_id is not None:
				left.player_id = right.player_id
		elif right.player_id is None:
			right.player_id = left.player_id
		else:
			assert left.player_id == right.player_id

		if left.name is None or left.name == UNKNOWN_HUMAN_PLAYER:
			if right.name is not None:
				left.name = right.name
		elif right.player_id is None or right.name == UNKNOWN_HUMAN_PLAYER:
			if left.name is not None:
				right.name = left.name

	def notify_first_player(self, entity_id: int):
		if entity_id in self._players_by_entity_id:
			first_player = self._players_by_entity_id[entity_id]
		else:
			first_player = PlayerReference(entity_id=entity_id)
			self._players_by_entity_id[entity_id] = first_player

		self.first_player = first_player

	def register_controller(self, entity_id: int, player_id: int):
		self._entity_controller_map[entity_id] = player_id

	def get_controller_by_entity_id(self, entity_id: int) -> Optional[int]:
		return self._entity_controller_map.get(entity_id)
